=== modified file 'NEWS'
--- NEWS	2008-05-10 03:13:56 +0000
+++ NEWS	2008-05-11 23:49:50 +0000
@@ -24,9 +24,14 @@
 
   INTERNALS:
 
-
   API BREAKS:
 
+    * ``VersionedFile.join`` is deprecated. This method required local
+      instances of both versioned file objects and was thus hostile to being
+      used for streaming from a smart server. The new get_record_stream and
+      insert_record_stream are meant to efficiently replace this method.
+      (Robert Collins)
+
 
 bzr 1.5rc1 2008-05-09
 ---------------------

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2008-05-07 10:04:52 +0000
+++ bzrlib/errors.py	2008-05-11 23:49:50 +0000
@@ -562,6 +562,18 @@
         PathError.__init__(self, base, reason)
 
 
+class UnavailableRepresentation(InternalBzrError):
+
+    _fmt = ("The encoding '%(wanted)s' is not available for key %(key)s which "
+        "is encoded as '%(native)s'.")
+
+    def __init__(self, key, wanted, native):
+        InternalBzrError.__init__(self)
+        self.wanted = wanted
+        self.native = native
+        self.key = key
+
+
 class UnknownHook(BzrError):
 
     _fmt = "The %(type)s hook '%(hook)s' is unknown in this version of bzrlib."

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2008-05-10 03:13:56 +0000
+++ bzrlib/fetch.py	2008-05-12 01:23:41 +0000
@@ -40,8 +40,7 @@
         )
 from bzrlib.trace import mutter
 import bzrlib.ui
-
-from bzrlib.lazy_import import lazy_import
+from bzrlib.versionedfile import filter_absent
 
 # TODO: Avoid repeatedly opening weaves so many times.
 
@@ -217,9 +216,9 @@
             self.to_repository.get_transaction())
         from_weave = self.from_weaves.get_weave(file_id,
             self.from_repository.get_transaction())
-        # we fetch all the texts, because texts do
-        # not reference anything, and its cheap enough
-        to_weave.join(from_weave, version_ids=required_versions)
+        # Fetch all the texts.
+        to_weave.insert_record_stream(from_weave.get_record_stream(
+            required_versions, 'topological', False))
 
     def _fetch_inventory_weave(self, revs, pb):
         pb.update("fetch inventory", 0, 2)
@@ -235,8 +234,8 @@
             # know for unselected inventories whether all their required
             # texts are present in the other repository - it could be
             # corrupt.
-            to_weave.join(from_weave, pb=child_pb, msg='merge inventory',
-                          version_ids=revs)
+            to_weave.insert_record_stream(from_weave.get_record_stream(revs,
+                'topological', False))
         finally:
             child_pb.finished()
 
@@ -301,7 +300,9 @@
             to_transaction)
         from_sf = self.from_repository._revision_store.get_signature_file(
             from_transaction)
-        to_sf.join(from_sf, version_ids=revs, ignore_missing=True)
+        # A missing signature is just skipped.
+        to_sf.insert_record_stream(filter_absent(from_sf.get_record_stream(revs,
+            'unordered', False)))
         self._fetch_just_revision_texts(revs, from_transaction, to_transaction)
 
     def _fetch_just_revision_texts(self, version_ids, from_transaction,
@@ -310,7 +311,8 @@
             to_transaction)
         from_rf = self.from_repository._revision_store.get_revision_file(
             from_transaction)
-        to_rf.join(from_rf, version_ids=version_ids)
+        to_rf.insert_record_stream(from_rf.get_record_stream(version_ids,
+            'topological', False))
 
 
 class Inter1and2Helper(object):

=== modified file 'bzrlib/knit.py'
--- bzrlib/knit.py	2008-05-08 04:33:38 +0000
+++ bzrlib/knit.py	2008-05-12 01:23:41 +0000
@@ -107,6 +107,7 @@
     contains_linebreaks,
     sha_string,
     sha_strings,
+    split_lines,
     )
 from bzrlib.symbol_versioning import (
     DEPRECATED_PARAMETER,
@@ -117,7 +118,13 @@
 from bzrlib.tsort import topo_sort
 from bzrlib.tuned_gzip import GzipFile, bytes_to_gzip
 import bzrlib.ui
-from bzrlib.versionedfile import VersionedFile, InterVersionedFile
+from bzrlib.versionedfile import (
+    AbsentContentFactory,
+    adapter_registry,
+    ContentFactory,
+    InterVersionedFile,
+    VersionedFile,
+    )
 import bzrlib.weave
 
 
@@ -138,6 +145,147 @@
 INDEX_SUFFIX = '.kndx'
 
 
+class KnitAdapter(object):
+    """Base class for knit record adaption."""
+
+    def __init__(self, basis_vf):
+        """Create an adapter which accesses full texts from basis_vf.
+        
+        :param basis_vf: A versioned file to access basis texts of deltas from.
+            May be None for adapters that do not need to access basis texts.
+        """
+        self._data = _KnitData(None)
+        self._annotate_factory = KnitAnnotateFactory()
+        self._plain_factory = KnitPlainFactory()
+        self._basis_vf = basis_vf
+
+
+class FTAnnotatedToUnannotated(KnitAdapter):
+    """An adapter from FT annotated knits to unannotated ones."""
+
+    def get_bytes(self, factory, annotated_compressed_bytes):
+        rec, contents = \
+            self._data._parse_record_unchecked(annotated_compressed_bytes)
+        content = self._annotate_factory.parse_fulltext(contents, rec[1])
+        size, bytes = self._data._record_to_data(rec[1], rec[3], content.text())
+        return bytes
+
+
+class DeltaAnnotatedToUnannotated(KnitAdapter):
+    """An adapter for deltas from annotated to unannotated."""
+
+    def get_bytes(self, factory, annotated_compressed_bytes):
+        rec, contents = \
+            self._data._parse_record_unchecked(annotated_compressed_bytes)
+        delta = self._annotate_factory.parse_line_delta(contents, rec[1],
+            plain=True)
+        contents = self._plain_factory.lower_line_delta(delta)
+        size, bytes = self._data._record_to_data(rec[1], rec[3], contents)
+        return bytes
+
+
+class FTAnnotatedToFullText(KnitAdapter):
+    """An adapter from FT annotated knits to unannotated ones."""
+
+    def get_bytes(self, factory, annotated_compressed_bytes):
+        rec, contents = \
+            self._data._parse_record_unchecked(annotated_compressed_bytes)
+        content, delta = self._annotate_factory.parse_record(factory.key[0],
+            contents, factory._build_details, None)
+        return ''.join(content.text())
+
+
+class DeltaAnnotatedToFullText(KnitAdapter):
+    """An adapter for deltas from annotated to unannotated."""
+
+    def get_bytes(self, factory, annotated_compressed_bytes):
+        rec, contents = \
+            self._data._parse_record_unchecked(annotated_compressed_bytes)
+        delta = self._annotate_factory.parse_line_delta(contents, rec[1],
+            plain=True)
+        compression_parent = factory.parents[0][0]
+        basis_lines = self._basis_vf.get_lines(compression_parent)
+        # Manually apply the delta because we have one annotated content and
+        # one plain.
+        basis_content = PlainKnitContent(basis_lines, compression_parent)
+        basis_content.apply_delta(delta, rec[1])
+        basis_content._should_strip_eol = factory._build_details[1]
+        return ''.join(basis_content.text())
+
+
+class FTPlainToFullText(KnitAdapter):
+    """An adapter from FT plain knits to unannotated ones."""
+
+    def get_bytes(self, factory, compressed_bytes):
+        rec, contents = \
+            self._data._parse_record_unchecked(compressed_bytes)
+        content, delta = self._plain_factory.parse_record(factory.key[0],
+            contents, factory._build_details, None)
+        return ''.join(content.text())
+
+
+class DeltaPlainToFullText(KnitAdapter):
+    """An adapter for deltas from annotated to unannotated."""
+
+    def get_bytes(self, factory, compressed_bytes):
+        rec, contents = \
+            self._data._parse_record_unchecked(compressed_bytes)
+        delta = self._plain_factory.parse_line_delta(contents, rec[1])
+        compression_parent = factory.parents[0][0]
+        basis_lines = self._basis_vf.get_lines(compression_parent)
+        basis_content = PlainKnitContent(basis_lines, compression_parent)
+        # Manually apply the delta because we have one annotated content and
+        # one plain.
+        content, _ = self._plain_factory.parse_record(rec[1], contents,
+            factory._build_details, basis_content)
+        return ''.join(content.text())
+
+
+class KnitContentFactory(ContentFactory):
+    """Content factory for streaming from knits.
+    
+    :seealso ContentFactory:
+    """
+
+    def __init__(self, version, parents, build_details, sha1, raw_record,
+        annotated, knit=None):
+        """Create a KnitContentFactory for version.
+        
+        :param version: The version.
+        :param parents: The parents.
+        :param build_details: The build details as returned from
+            get_build_details.
+        :param sha1: The sha1 expected from the full text of this object.
+        :param raw_record: The bytes of the knit data from disk.
+        :param annotated: True if the raw data is annotated.
+        """
+        ContentFactory.__init__(self)
+        self.sha1 = sha1
+        self.key = (version,)
+        self.parents = tuple((parent,) for parent in parents)
+        if build_details[0] == 'line-delta':
+            kind = 'delta'
+        else:
+            kind = 'ft'
+        if annotated:
+            annotated_kind = 'annotated-'
+        else:
+            annotated_kind = ''
+        self.storage_kind = 'knit-%s%s-gz' % (annotated_kind, kind)
+        self._raw_record = raw_record
+        self._build_details = build_details
+        self._knit = knit
+
+    def get_bytes_as(self, storage_kind):
+        if storage_kind == self.storage_kind:
+            return self._raw_record
+        if storage_kind == 'fulltext' and self._knit is not None:
+            return self._knit.get_text(self.key[0])
+        else:
+            raise errors.UnavailableRepresentation(self.key, storage_kind,
+                self.storage_kind)
+
+
 class KnitContent(object):
     """Content of a knit version to which deltas can be applied."""
 
@@ -231,8 +379,7 @@
                 % (e,))
 
         if self._should_strip_eol:
-            anno, line = lines[-1]
-            lines[-1] = (anno, line.rstrip('\n'))
+            lines[-1] = lines[-1].rstrip('\n')
         return lines
 
     def copy(self):
@@ -499,8 +646,6 @@
     """Factory to create a KnitVersionedFile for a .knit/.kndx file pair."""
     if factory is None:
         factory = KnitAnnotateFactory()
-    else:
-        factory = KnitPlainFactory()
     if get_scope is None:
         get_scope = lambda:None
     index = _KnitIndex(transport, name + INDEX_SUFFIX,
@@ -612,12 +757,10 @@
         # write all the data
         raw_record_sizes = [record[3] for record in records]
         positions = self._data.add_raw_records(raw_record_sizes, data)
-        offset = 0
         index_entries = []
-        for (version_id, options, parents, size), access_memo in zip(
+        for (version_id, options, parents, _), access_memo in zip(
             records, positions):
             index_entries.append((version_id, options, access_memo, parents))
-            offset += size
         self._index.add_versions(index_entries)
 
     def copy_to(self, name, transport):
@@ -698,7 +841,7 @@
         # know the length of all the data a-priori.
         raw_datum = []
         result_version_list = []
-        for (version_id, raw_data), \
+        for (version_id, raw_data, _), \
             (version_id2, options, _, parents) in \
             izip(self._data.read_records_iter_raw(copy_queue_records),
                  temp_version_list):
@@ -716,6 +859,53 @@
                 return pseudo_file.read(length)
         return (self.get_format_signature(), result_version_list, read)
 
+    def get_record_stream(self, versions, ordering, include_delta_closure):
+        """Get a stream of records for versions.
+
+        :param versions: The versions to include. Each version is a tuple
+            (version,).
+        :param ordering: Either 'unordered' or 'topological'. A topologically
+            sorted stream has compression parents strictly before their
+            children.
+        :param include_delta_closure: If True then the closure across any
+            compression parents will be included (in the opaque data).
+        :return: An iterator of ContentFactory objects, each of which is only
+            valid until the iterator is advanced.
+        """
+        if include_delta_closure:
+            # Nb: what we should do is plan the data to stream to allow
+            # reconstruction of all the texts without excessive buffering,
+            # including re-sending common bases as needed. This makes the most
+            # sense when we start serialising these streams though, so for now
+            # we just fallback to individual text construction behind the
+            # abstraction barrier.
+            knit = self
+        else:
+            knit = None
+        # We end up doing multiple index lookups here for parents details and
+        # disk layout details - we need a unified api ?
+        parent_map = self.get_parent_map(versions)
+        absent_versions = set(versions) - set(parent_map)
+        if ordering == 'topological':
+            present_versions = topo_sort(parent_map)
+        else:
+            # List comprehension to keep the requested order (as that seems
+            # marginally useful, at least until we start doing IO optimising
+            # here.
+            present_versions = [version for version in versions if version in
+                parent_map]
+        position_map = self._get_components_positions(present_versions)
+        records = [(version, position_map[version][1]) for version in
+            present_versions]
+        record_map = {}
+        for version in absent_versions:
+            yield AbsentContentFactory((version,))
+        for version, raw_data, sha1 in \
+                self._data.read_records_iter_raw(records):
+            (record_details, index_memo, _) = position_map[version]
+            yield KnitContentFactory(version, parent_map[version],
+                record_details, sha1, raw_data, self.factory.annotated, knit)
+
     def _extract_blocks(self, version_id, source, target):
         if self._index.get_method(version_id) != 'line-delta':
             return None
@@ -797,7 +987,8 @@
                     'incompatible format signature inserting to %r', self)
             source = self._knit_from_datastream(
                 (format, data_list, reader_callable))
-            self.join(source)
+            stream = source.get_record_stream(source.versions(), 'unordered', False)
+            self.insert_record_stream(stream)
             return
 
         for version_id, options, length, parents in data_list:
@@ -886,6 +1077,112 @@
         return KnitVersionedFile(self.filename, self.transport,
             factory=factory, index=index, access_method=access)
 
+    def insert_record_stream(self, stream):
+        """Insert a record stream into this versioned file.
+
+        :param stream: A stream of records to insert. 
+        :return: None
+        :seealso VersionedFile.get_record_stream:
+        """
+        def get_adapter(adapter_key):
+            try:
+                return adapters[adapter_key]
+            except KeyError:
+                adapter_factory = adapter_registry.get(adapter_key)
+                adapter = adapter_factory(self)
+                adapters[adapter_key] = adapter
+                return adapter
+        if self.factory.annotated:
+            # self is annotated, we need annotated knits to use directly.
+            annotated = "annotated-"
+            convertibles = []
+        else:
+            # self is not annotated, but we can strip annotations cheaply.
+            annotated = ""
+            convertibles = set(["knit-annotated-delta-gz",
+                "knit-annotated-ft-gz"])
+        # The set of types we can cheaply adapt without needing basis texts.
+        native_types = set()
+        native_types.add("knit-%sdelta-gz" % annotated)
+        native_types.add("knit-%sft-gz" % annotated)
+        knit_types = native_types.union(convertibles)
+        adapters = {}
+        # Buffer all index entries that we can't add immediately because their
+        # basis parent is missing. We don't buffer all because generating
+        # annotations may require access to some of the new records. However we
+        # can't generate annotations from new deltas until their basis parent
+        # is present anyway, so we get away with not needing an index that
+        # includes the new keys.
+        # key = basis_parent, value = index entry to add
+        buffered_index_entries = {}
+        for record in stream:
+            # Raise an error when a record is missing.
+            if record.storage_kind == 'absent':
+                raise RevisionNotPresent([record.key[0]], self)
+            # adapt to non-tuple interface
+            parents = [parent[0] for parent in record.parents]
+            if record.storage_kind in knit_types:
+                if record.storage_kind not in native_types:
+                    try:
+                        adapter_key = (record.storage_kind, "knit-delta-gz")
+                        adapter = get_adapter(adapter_key)
+                    except KeyError:
+                        adapter_key = (record.storage_kind, "knit-ft-gz")
+                        adapter = get_adapter(adapter_key)
+                    bytes = adapter.get_bytes(
+                        record, record.get_bytes_as(record.storage_kind))
+                else:
+                    bytes = record.get_bytes_as(record.storage_kind)
+                options = [record._build_details[0]]
+                if record._build_details[1]:
+                    options.append('no-eol')
+                # Just blat it across.
+                # Note: This does end up adding data on duplicate keys. As
+                # modern repositories use atomic insertions this should not
+                # lead to excessive growth in the event of interrupted fetches.
+                # 'knit' repositories may suffer excessive growth, but as a
+                # deprecated format this is tolerable. It can be fixed if
+                # needed by in the kndx index support raising on a duplicate
+                # add with identical parents and options.
+                access_memo = self._data.add_raw_records([len(bytes)], bytes)[0]
+                index_entry = (record.key[0], options, access_memo, parents)
+                buffered = False
+                if 'fulltext' not in options:
+                    basis_parent = parents[0]
+                    if not self.has_version(basis_parent):
+                        pending = buffered_index_entries.setdefault(
+                            basis_parent, [])
+                        pending.append(index_entry)
+                        buffered = True
+                if not buffered:
+                    self._index.add_versions([index_entry])
+            elif record.storage_kind == 'fulltext':
+                self.add_lines(record.key[0], parents,
+                    split_lines(record.get_bytes_as('fulltext')))
+            else:
+                adapter_key = record.storage_kind, 'fulltext'
+                adapter = get_adapter(adapter_key)
+                lines = split_lines(adapter.get_bytes(
+                    record, record.get_bytes_as(record.storage_kind)))
+                try:
+                    self.add_lines(record.key[0], parents, lines)
+                except errors.RevisionAlreadyPresent:
+                    pass
+            # Add any records whose basis parent is now available.
+            added_keys = [record.key[0]]
+            while added_keys:
+                key = added_keys.pop(0)
+                if key in buffered_index_entries:
+                    index_entries = buffered_index_entries[key]
+                    self._index.add_versions(index_entries)
+                    added_keys.extend(
+                        [index_entry[0] for index_entry in index_entries])
+                    del buffered_index_entries[key]
+        # If there were any deltas which had a missing basis parent, error.
+        if buffered_index_entries:
+            raise errors.RevisionNotPresent(buffered_index_entries.keys()[0],
+                self)
+
     def versions(self):
         """See VersionedFile.versions."""
         if 'evil' in debug.debug_flags:
@@ -1101,6 +1398,19 @@
 
     def check(self, progress_bar=None):
         """See VersionedFile.check()."""
+        # This doesn't actually test extraction of everything, but that will
+        # impact 'bzr check' substantially, and needs to be integrated with
+        # care. However, it does check for the obvious problem of a delta with
+        # no basis.
+        versions = self.versions()
+        parent_map = self.get_parent_map(versions)
+        for version in versions:
+            if self._index.get_method(version) != 'fulltext':
+                compression_parent = parent_map[version][0]
+                if compression_parent not in parent_map:
+                    raise errors.KnitCorrupt(self,
+                        "Missing basis parent %s for %s" % (
+                        compression_parent, version))
 
     def get_lines(self, version_id):
         """See VersionedFile.get_lines()."""
@@ -2441,45 +2751,54 @@
                               % (version_id, e.__class__.__name__, str(e)))
         return df, rec
 
-    def _check_header(self, version_id, line):
+    def _split_header(self, line):
         rec = line.split()
         if len(rec) != 4:
             raise KnitCorrupt(self._access,
                               'unexpected number of elements in record header')
+        return rec
+
+    def _check_header_version(self, rec, version_id):
         if rec[1] != version_id:
             raise KnitCorrupt(self._access,
                               'unexpected version, wanted %r, got %r'
                               % (version_id, rec[1]))
+
+    def _check_header(self, version_id, line):
+        rec = self._split_header(line)
+        self._check_header_version(rec, version_id)
         return rec
 
-    def _parse_record(self, version_id, data):
+    def _parse_record_unchecked(self, data):
         # profiling notes:
         # 4168 calls in 2880 217 internal
         # 4168 calls to _parse_record_header in 2121
         # 4168 calls to readlines in 330
         df = GzipFile(mode='rb', fileobj=StringIO(data))
-
         try:
             record_contents = df.readlines()
         except Exception, e:
-            raise KnitCorrupt(self._access,
-                              "While reading {%s} got %s(%s)"
-                              % (version_id, e.__class__.__name__, str(e)))
+            raise KnitCorrupt(self._access, "Corrupt compressed record %r, got %s(%s)" %
+                (data, e.__class__.__name__, str(e)))
         header = record_contents.pop(0)
-        rec = self._check_header(version_id, header)
-
+        rec = self._split_header(header)
         last_line = record_contents.pop()
         if len(record_contents) != int(rec[2]):
             raise KnitCorrupt(self._access,
                               'incorrect number of lines %s != %s'
                               ' for version {%s}'
                               % (len(record_contents), int(rec[2]),
-                                 version_id))
+                                 rec[1]))
         if last_line != 'end %s\n' % rec[1]:
             raise KnitCorrupt(self._access,
                               'unexpected version end line %r, wanted %r' 
-                              % (last_line, version_id))
+                              % (last_line, rec[1]))
         df.close()
+        return rec, record_contents
+
+    def _parse_record(self, version_id, data):
+        rec, record_contents = self._parse_record_unchecked(data)
+        self._check_header_version(rec, version_id)
         return record_contents, rec[3]
 
     def read_records_iter_raw(self, records):
@@ -2487,6 +2806,9 @@
 
         This unpacks enough of the text record to validate the id is
         as expected but thats all.
+
+        Each item the iterator yields is (version_id, bytes,
+        sha1_of_full_text).
         """
         # setup an iterator of the external records:
         # uses readv so nice and fast we hope.
@@ -2501,7 +2823,7 @@
             # validate the header
             df, rec = self._parse_record_header(version_id, data)
             df.close()
-            yield version_id, data
+            yield version_id, data, rec[3]
 
     def read_records_iter(self, records):
         """Read text records from data file and yield result.
@@ -2666,7 +2988,7 @@
             total = len(version_list)
             raw_datum = []
             raw_records = []
-            for (version_id, raw_data), \
+            for (version_id, raw_data, _), \
                 (version_id2, options, parents) in \
                 izip(self.source._data.read_records_iter_raw(copy_queue_records),
                      copy_queue):

=== modified file 'bzrlib/store/__init__.py'
--- bzrlib/store/__init__.py	2008-04-24 07:22:53 +0000
+++ bzrlib/store/__init__.py	2008-05-11 23:49:50 +0000
@@ -73,10 +73,6 @@
         """DEPRECATED. Please use .get(fileid) instead."""
         raise NotImplementedError
 
-    #def __contains__(self, fileid):
-    #    """Deprecated, please use has_id"""
-    #    raise NotImplementedError
-
     def __iter__(self):
         raise NotImplementedError
 

=== modified file 'bzrlib/store/versioned/__init__.py'
--- bzrlib/store/versioned/__init__.py	2008-05-10 03:13:56 +0000
+++ bzrlib/store/versioned/__init__.py	2008-05-11 23:49:50 +0000
@@ -178,7 +178,8 @@
     def _put_weave(self, file_id, weave, transaction):
         """Preserved here for upgrades-to-weaves to use."""
         myweave = self._make_new_versionedfile(file_id, transaction)
-        myweave.join(weave)
+        myweave.insert_record_stream(weave.get_record_stream(weave.versions(),
+            'topological', False))
 
     def copy(self, source, result_id, transaction):
         """Copy the source versioned file to result_id in this store."""
@@ -238,7 +239,9 @@
                 # joining is fast with knits, and bearable for weaves -
                 # indeed the new case can be optimised if needed.
                 target = self._make_new_versionedfile(f, to_transaction)
-                target.join(from_store.get_weave(f, from_transaction))
+                source = from_store.get_weave(f, from_transaction)
+                target.insert_record_stream(source.get_record_stream(
+                    source.versions(), 'topological', False))
         finally:
             pb.finished()
 

=== modified file 'bzrlib/symbol_versioning.py'
--- bzrlib/symbol_versioning.py	2008-05-10 11:33:32 +0000
+++ bzrlib/symbol_versioning.py	2008-05-11 23:49:50 +0000
@@ -93,6 +93,7 @@
     return ("%s was deprecated in version "
             + bzrlib._format_version_tuple(version_tuple))
 
+
 def set_warning_method(method):
     """Set the warning method to be used by this module.
 

=== modified file 'bzrlib/tests/interrepository_implementations/test_fetch.py'
--- bzrlib/tests/interrepository_implementations/test_fetch.py	2008-04-27 23:47:45 +0000
+++ bzrlib/tests/interrepository_implementations/test_fetch.py	2008-05-11 23:49:50 +0000
@@ -93,23 +93,23 @@
         finally:
             to_repo.unlock()
 
-        # Implementations can either copy the missing basis text, or raise an
-        # exception
+        # Implementations can either ensure that the target of the delta is
+        # reconstructable, or raise an exception (which stream based copies
+        # generally do).
         try:
             to_repo.fetch(tree.branch.repository, 'rev-two')
         except errors.RevisionNotPresent, e:
             # If an exception is raised, the revision should not be in the
             # target.
-            self.assertRaises(
-                (errors.NoSuchRevision, errors.RevisionNotPresent),
-                 to_repo.revision_tree, 'rev-two')
+            self.assertRaises((errors.NoSuchRevision, errors.RevisionNotPresent),
+                              to_repo.revision_tree, 'rev-two')
         else:
-            # If not exception is raised, then the basis text should be
+            # If not exception is raised, then the text should be
             # available.
             to_repo.lock_read()
             try:
-                rt = to_repo.revision_tree('rev-one')
-                self.assertEqual('contents of tree/a\n',
+                rt = to_repo.revision_tree('rev-two')
+                self.assertEqual('new contents\n',
                                  rt.get_file_text('a-id'))
             finally:
                 to_repo.unlock()
@@ -118,8 +118,7 @@
         repo_a = self.make_repository('.')
         repo_b = repository.Repository.open('.')
         try:
-            self.assertRaises(errors.NoSuchRevision, repo_b.fetch, repo_a,
-                              revision_id='XXX')
+            self.assertRaises(errors.NoSuchRevision, repo_b.fetch, repo_a, revision_id='XXX')
         except errors.LockError, e:
             check_old_format_lock_error(self.repository_format)
 
@@ -173,8 +172,7 @@
         from_tree.add('filename', 'funky-chars<>%&;"\'')
         from_tree.commit('commit filename')
         to_repo = self.make_to_repository('to')
-        to_repo.fetch(from_tree.branch.repository,
-                      from_tree.get_parent_ids()[0])
+        to_repo.fetch(from_tree.branch.repository, from_tree.get_parent_ids()[0])
 
     def test_fetch_revision_hash(self):
         """Ensure that inventory hashes are updated by fetch"""

=== modified file 'bzrlib/tests/interrepository_implementations/test_interrepository.py'
--- bzrlib/tests/interrepository_implementations/test_interrepository.py	2008-04-27 22:35:42 +0000
+++ bzrlib/tests/interrepository_implementations/test_interrepository.py	2008-05-11 23:49:50 +0000
@@ -91,7 +91,6 @@
         pass
 
 
-
 class TestCaseWithComplexRepository(TestCaseWithInterRepository):
 
     def setUp(self):

=== modified file 'bzrlib/tests/interversionedfile_implementations/test_join.py'
--- bzrlib/tests/interversionedfile_implementations/test_join.py	2008-03-27 04:54:12 +0000
+++ bzrlib/tests/interversionedfile_implementations/test_join.py	2008-04-29 05:27:08 +0000
@@ -18,6 +18,7 @@
 
 
 import bzrlib.errors as errors
+from bzrlib.symbol_versioning import one_five
 from bzrlib.tests import TestCaseWithTransport
 from bzrlib.transport import get_transport
 import bzrlib.versionedfile as versionedfile
@@ -44,7 +45,7 @@
         f1.add_lines('r0', [], ['a\n', 'b\n'])
         f1.add_lines('r1', ['r0'], ['c\n', 'b\n'])
         f2 = self.get_target()
-        f2.join(f1, None)
+        self.applyDeprecated(one_five, f2.join, f1, None)
         def verify_file(f):
             self.assertTrue(f.has_version('r0'))
             self.assertTrue(f.has_version('r1'))
@@ -52,7 +53,7 @@
         verify_file(self.get_target())
 
         self.assertRaises(errors.RevisionNotPresent,
-            f2.join, f1, version_ids=['r3'])
+            self.applyDeprecated, one_five, f2.join, f1, version_ids=['r3'])
 
     def test_gets_expected_inter_worker(self):
         source = self.get_source()
@@ -72,7 +73,8 @@
         source.add_lines('ancestorright', ['base'], [])
         source.add_lines('namedleft', ['ancestorleft'], [])
         source.add_lines('namedright', ['ancestorright'], [])
-        target.join(source, version_ids=['namedleft', 'namedright'])
+        self.applyDeprecated(one_five, target.join, source,
+            version_ids=['namedleft', 'namedright'])
         self.assertFalse(target.has_version('sibling'))
         self.assertTrue(target.has_version('ancestorleft'))
         self.assertTrue(target.has_version('ancestorright'))
@@ -88,7 +90,7 @@
         w1.add_lines('v-3', ['v-1'], ['line 1\n'])
         w2.add_lines('v-3', ['v-2'], ['line 1\n'])
         try:
-            w1.join(w2)
+            self.applyDeprecated(one_five, w1.join, w2)
         except errors.WeaveParentMismatch:
             # Acceptable behaviour:
             return
@@ -115,7 +117,7 @@
         """Reweave adding empty weave"""
         wb = self.get_target()
         w1 = self.build_weave1()
-        w1.join(wb)
+        self.applyDeprecated(one_five, w1.join, wb)
         self.verify_weave1(w1)
 
     def verify_weave1(self, w1):
@@ -136,7 +138,7 @@
         t = self.get_target()
         t.add_lines('base', [], [])
         t.add_lines('text', ['base'], [])
-        t.join(s)
+        self.applyDeprecated(one_five, t.join, s)
         self.assertEqual({'text':('base',)}, t.get_parent_map(['text']))
 
     def test_join_with_ghosts(self):
@@ -159,7 +161,7 @@
         wb.add_lines('v1', [], ['hello\n'])
         wb.add_lines('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
         try:
-            w1.join(wb)
+            self.applyDeprecated(one_five, w1.join, wb)
         except errors.WeaveParentMismatch:
             # Acceptable behaviour:
             return
@@ -176,7 +178,8 @@
         wb.add_lines('x1', [], ['line from x1\n'])
         wb.add_lines('v1', [], ['hello\n'])
         wb.add_lines('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
-        w1.join(wb, version_ids=['x1', 'z1'], ignore_missing=True)
+        self.applyDeprecated(one_five, w1.join, wb, version_ids=['x1', 'z1'],
+            ignore_missing=True)
         eq = self.assertEquals
         eq(sorted(w1.versions()), ['v1', 'v2', 'v3', 'x1'])
         eq(w1.get_text('x1'), 'line from x1\n')
@@ -213,7 +216,7 @@
             # The target does not support ghosts; the test is irrelevant.
             return
         try:
-            source.join(target)
+            self.applyDeprecated(one_five, source.join, target)
         except errors.RevisionNotPresent:
             return
         # legacy apis should behave
@@ -228,7 +231,7 @@
         # if we add something that is fills out what is a ghost, then 
         # when joining into a ghost aware join it should flesh out the ghosts.
         source.add_lines('base', [], [])
-        target.join(source, version_ids=['base'])
+        self.applyDeprecated(one_five, target.join, source, version_ids=['base'])
         self.assertEqual(['base', 'notbase'], target.get_ancestry(['notbase']))
         self.assertEqual({'base':(),
                           'notbase':('base', ),
@@ -251,7 +254,7 @@
         source.add_lines('inherit_me', [], ['b\n'])
         source.add_lines('select_me', ['inherit_me'], ['b\n'])
         target = self.get_target()
-        target.join(source, version_ids=['select_me'])
+        self.applyDeprecated(one_five, target.join, source, version_ids=['select_me'])
         self.assertEqual(['inherit_me', 'select_me'], target.versions())
 
     def test_join_odd_records(self):
@@ -267,7 +270,7 @@
         source.add_lines('4', ['2'], ['1st\n'])
         source.add_lines('5', ['3'], ['1st\n', '2nd\n', '3rd\n'])
         target = self.get_target()
-        target.join(source, version_ids=['1', '3', '5'])
+        self.applyDeprecated(one_five, target.join, source, version_ids=['1', '3', '5'])
         target = self.get_target(create=False)
         self.assertEqual(set(['1', '3', '5']), set(target.versions()))
         self.assertEqual(3, len(target.versions()))

=== modified file 'bzrlib/tests/test_errors.py'
--- bzrlib/tests/test_errors.py	2008-05-02 07:31:24 +0000
+++ bzrlib/tests/test_errors.py	2008-05-11 23:49:50 +0000
@@ -199,6 +199,12 @@
             "the currently open request.",
             str(error))
 
+    def test_unavailable_representation(self):
+        error = errors.UnavailableRepresentation(('key',), "mpdiff", "fulltext")
+        self.assertEqualDiff("The encoding 'mpdiff' is not available for key "
+            "('key',) which is encoded as 'fulltext'.",
+            str(error))
+
     def test_unknown_hook(self):
         error = errors.UnknownHook("branch", "foo")
         self.assertEqualDiff("The branch hook 'foo' is unknown in this version"

=== modified file 'bzrlib/tests/test_knit.py'
--- bzrlib/tests/test_knit.py	2008-04-28 06:24:16 +0000
+++ bzrlib/tests/test_knit.py	2008-05-11 23:49:50 +0000
@@ -370,7 +370,7 @@
         self.assertEqual({'rev-id-1':(['foo\n', 'bar\n'], sha1sum)}, contents)
 
         raw_contents = list(data.read_records_iter_raw(records))
-        self.assertEqual([('rev-id-1', gz_txt)], raw_contents)
+        self.assertEqual([('rev-id-1', gz_txt, sha1sum)], raw_contents)
 
     def test_not_enough_lines(self):
         sha1sum = sha.new('foo\n').hexdigest()
@@ -387,7 +387,7 @@
 
         # read_records_iter_raw won't detect that sort of mismatch/corruption
         raw_contents = list(data.read_records_iter_raw(records))
-        self.assertEqual([('rev-id-1', gz_txt)], raw_contents)
+        self.assertEqual([('rev-id-1', gz_txt, sha1sum)], raw_contents)
 
     def test_too_many_lines(self):
         sha1sum = sha.new('foo\nbar\n').hexdigest()
@@ -405,7 +405,7 @@
 
         # read_records_iter_raw won't detect that sort of mismatch/corruption
         raw_contents = list(data.read_records_iter_raw(records))
-        self.assertEqual([('rev-id-1', gz_txt)], raw_contents)
+        self.assertEqual([('rev-id-1', gz_txt, sha1sum)], raw_contents)
 
     def test_mismatched_version_id(self):
         sha1sum = sha.new('foo\nbar\n').hexdigest()
@@ -1056,7 +1056,7 @@
         """
         index_memo = knit._index.get_position(version_id)
         record = (version_id, index_memo)
-        [(_, expected_content)] = list(knit._data.read_records_iter_raw([record]))
+        [(_, expected_content, _)] = list(knit._data.read_records_iter_raw([record]))
         self.assertEqual(expected_content, candidate_content)
 
 
@@ -1293,39 +1293,6 @@
         self.assertEquals(origins[1], ('text-1', 'b\n'))
         self.assertEquals(origins[2], ('text-1', 'c\n'))
 
-    def _test_join_with_factories(self, k1_factory, k2_factory):
-        k1 = make_file_knit('test1', get_transport('.'), factory=k1_factory, create=True)
-        k1.add_lines('text-a', [], ['a1\n', 'a2\n', 'a3\n'])
-        k1.add_lines('text-b', ['text-a'], ['a1\n', 'b2\n', 'a3\n'])
-        k1.add_lines('text-c', [], ['c1\n', 'c2\n', 'c3\n'])
-        k1.add_lines('text-d', ['text-c'], ['c1\n', 'd2\n', 'd3\n'])
-        k1.add_lines('text-m', ['text-b', 'text-d'], ['a1\n', 'b2\n', 'd3\n'])
-        k2 = make_file_knit('test2', get_transport('.'), factory=k2_factory, create=True)
-        count = k2.join(k1, version_ids=['text-m'])
-        self.assertEquals(count, 5)
-        self.assertTrue(k2.has_version('text-a'))
-        self.assertTrue(k2.has_version('text-c'))
-        origins = k2.annotate('text-m')
-        self.assertEquals(origins[0], ('text-a', 'a1\n'))
-        self.assertEquals(origins[1], ('text-b', 'b2\n'))
-        self.assertEquals(origins[2], ('text-d', 'd3\n'))
-
-    def test_knit_join_plain_to_plain(self):
-        """Test joining a plain knit with a plain knit."""
-        self._test_join_with_factories(KnitPlainFactory(), KnitPlainFactory())
-
-    def test_knit_join_anno_to_anno(self):
-        """Test joining an annotated knit with an annotated knit."""
-        self._test_join_with_factories(None, None)
-
-    def test_knit_join_anno_to_plain(self):
-        """Test joining an annotated knit with a plain knit."""
-        self._test_join_with_factories(None, KnitPlainFactory())
-
-    def test_knit_join_plain_to_anno(self):
-        """Test joining a plain knit with an annotated knit."""
-        self._test_join_with_factories(KnitPlainFactory(), None)
-
     def test_reannotate(self):
         k1 = make_file_knit('knit1', get_transport('.'),
                                factory=KnitAnnotateFactory(), create=True)
@@ -1336,7 +1303,8 @@
 
         k2 = make_file_knit('test2', get_transport('.'),
                                factory=KnitAnnotateFactory(), create=True)
-        k2.join(k1, version_ids=['text-b'])
+        k2.insert_record_stream(k1.get_record_stream(k1.versions(),
+            'unordered', False))
 
         # 2
         k1.add_lines('text-X', ['text-b'], ['a\n', 'b\n'])
@@ -1346,7 +1314,8 @@
         k2.add_lines('text-Y', ['text-b'], ['b\n', 'c\n'])
 
         # test-c will have index 3
-        k1.join(k2, version_ids=['text-c'])
+        k1.insert_record_stream(k2.get_record_stream(['text-c'],
+            'unordered', False))
 
         lines = k1.get_lines('text-c')
         self.assertEquals(lines, ['z\n', 'c\n'])
@@ -2817,7 +2786,8 @@
         target_knit = self.make_test_knit(name='annotated', annotate=True)
         source_knit.add_lines("A", [], ["Foo\n"])
         # Give the target A, so we can try to thunk across to it.
-        target_knit.join(source_knit)
+        target_knit.insert_record_stream(source_knit.get_record_stream(['A'],
+            'unordered', False))
         index, access = self.get_index_access(target_knit,
             source_knit.get_data_stream([]))
         raw_data = list(access.get_raw_records([(True, "A", None, None)]))[0]

=== modified file 'bzrlib/tests/test_versionedfile.py'
--- bzrlib/tests/test_versionedfile.py	2008-04-08 21:41:15 +0000
+++ bzrlib/tests/test_versionedfile.py	2008-05-12 01:23:41 +0000
@@ -21,6 +21,7 @@
 # TODO: might be nice to create a versionedfile with some type of corruption
 # considered typical and check that it can be detected/corrected.
 
+from itertools import chain
 from StringIO import StringIO
 
 import bzrlib
@@ -34,23 +35,53 @@
                            RevisionAlreadyPresent,
                            WeaveParentMismatch
                            )
+from bzrlib import knit as _mod_knit
 from bzrlib.knit import (
     make_file_knit,
     KnitAnnotateFactory,
     KnitPlainFactory,
     )
-from bzrlib.symbol_versioning import one_four
+from bzrlib.symbol_versioning import one_four, one_five
 from bzrlib.tests import TestCaseWithMemoryTransport, TestSkipped
 from bzrlib.tests.http_utils import TestCaseWithWebserver
 from bzrlib.trace import mutter
 from bzrlib.transport import get_transport
 from bzrlib.transport.memory import MemoryTransport
 from bzrlib.tsort import topo_sort
+from bzrlib.tuned_gzip import GzipFile
 import bzrlib.versionedfile as versionedfile
 from bzrlib.weave import WeaveFile
 from bzrlib.weavefile import read_weave, write_weave
 
 
+def get_diamond_vf(f, trailing_eol=True, left_only=False):
+    """Get a diamond graph to exercise deltas and merges.
+    
+    :param trailing_eol: If True end the last line with \n.
+    """
+    parents = {
+        'origin': (),
+        'base': (('origin',),),
+        'left': (('base',),),
+        'right': (('base',),),
+        'merged': (('left',), ('right',)),
+        }
+    # insert a diamond graph to exercise deltas and merges.
+    if trailing_eol:
+        last_char = '\n'
+    else:
+        last_char = ''
+    f.add_lines('origin', [], ['origin' + last_char])
+    f.add_lines('base', ['origin'], ['base' + last_char])
+    f.add_lines('left', ['base'], ['base\n', 'left' + last_char])
+    if not left_only:
+        f.add_lines('right', ['base'],
+            ['base\n', 'right' + last_char])
+        f.add_lines('merged', ['left', 'right'],
+            ['base\n', 'left\n', 'right\n', 'merged' + last_char])
+    return f, parents
+
+
 class VersionedFileTestMixIn(object):
     """A mixin test class for testing VersionedFiles.
 
@@ -87,6 +118,266 @@
         f = self.reopen_file(create=True)
         verify_file(f)
 
+    def test_get_record_stream_empty(self):
+        """get_record_stream is a replacement for get_data_stream."""
+        f = self.get_file()
+        entries = f.get_record_stream([], 'unordered', False)
+        self.assertEqual([], list(entries))
+
+    def assertValidStorageKind(self, storage_kind):
+        """Assert that storage_kind is a valid storage_kind."""
+        self.assertSubset([storage_kind],
+            ['mpdiff', 'knit-annotated-ft', 'knit-annotated-delta',
+             'knit-ft', 'knit-delta', 'fulltext', 'knit-annotated-ft-gz',
+             'knit-annotated-delta-gz', 'knit-ft-gz', 'knit-delta-gz'])
+
+    def capture_stream(self, f, entries, on_seen, parents):
+        """Capture a stream for testing."""
+        for factory in entries:
+            on_seen(factory.key)
+            self.assertValidStorageKind(factory.storage_kind)
+            self.assertEqual(f.get_sha1s([factory.key[0]])[0], factory.sha1)
+            self.assertEqual(parents[factory.key[0]], factory.parents)
+            self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
+                str)
+
+    def test_get_record_stream_interface(self):
+        """Each item in a stream has to provide a regular interface."""
+        f, parents = get_diamond_vf(self.get_file())
+        entries = f.get_record_stream(['merged', 'left', 'right', 'base'],
+            'unordered', False)
+        seen = set()
+        self.capture_stream(f, entries, seen.add, parents)
+        self.assertEqual(set([('base',), ('left',), ('right',), ('merged',)]),
+            seen)
+
+    def test_get_record_stream_interface_ordered(self):
+        """Each item in a stream has to provide a regular interface."""
+        f, parents = get_diamond_vf(self.get_file())
+        entries = f.get_record_stream(['merged', 'left', 'right', 'base'],
+            'topological', False)
+        seen = []
+        self.capture_stream(f, entries, seen.append, parents)
+        self.assertSubset([tuple(seen)],
+            (
+             (('base',), ('left',), ('right',), ('merged',)),
+             (('base',), ('right',), ('left',), ('merged',)),
+            ))
+
+    def test_get_record_stream_interface_ordered_with_delta_closure(self):
+        """Each item in a stream has to provide a regular interface."""
+        f, parents = get_diamond_vf(self.get_file())
+        entries = f.get_record_stream(['merged', 'left', 'right', 'base'],
+            'topological', True)
+        seen = []
+        for factory in entries:
+            seen.append(factory.key)
+            self.assertValidStorageKind(factory.storage_kind)
+            self.assertEqual(f.get_sha1s([factory.key[0]])[0], factory.sha1)
+            self.assertEqual(parents[factory.key[0]], factory.parents)
+            self.assertEqual(f.get_text(factory.key[0]),
+                factory.get_bytes_as('fulltext'))
+            self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
+                str)
+        self.assertSubset([tuple(seen)],
+            (
+             (('base',), ('left',), ('right',), ('merged',)),
+             (('base',), ('right',), ('left',), ('merged',)),
+            ))
+
+    def test_get_record_stream_unknown_storage_kind_raises(self):
+        """Asking for a storage kind that the stream cannot supply raises."""
+        f, parents = get_diamond_vf(self.get_file())
+        entries = f.get_record_stream(['merged', 'left', 'right', 'base'],
+            'unordered', False)
+        # We track the contents because we should be able to try, fail a
+        # particular kind and then ask for one that works and continue.
+        seen = set()
+        for factory in entries:
+            seen.add(factory.key)
+            self.assertValidStorageKind(factory.storage_kind)
+            self.assertEqual(f.get_sha1s([factory.key[0]])[0], factory.sha1)
+            self.assertEqual(parents[factory.key[0]], factory.parents)
+            # currently no stream emits mpdiff
+            self.assertRaises(errors.UnavailableRepresentation,
+                factory.get_bytes_as, 'mpdiff')
+            self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
+                str)
+        self.assertEqual(set([('base',), ('left',), ('right',), ('merged',)]),
+            seen)
+
+    def test_get_record_stream_missing_records_are_absent(self):
+        f, parents = get_diamond_vf(self.get_file())
+        entries = f.get_record_stream(['merged', 'left', 'right', 'or', 'base'],
+            'unordered', False)
+        self.assertAbsentRecord(f, parents, entries)
+        entries = f.get_record_stream(['merged', 'left', 'right', 'or', 'base'],
+            'topological', False)
+        self.assertAbsentRecord(f, parents, entries)
+
+    def assertAbsentRecord(self, f, parents, entries):
+        """Helper for test_get_record_stream_missing_records_are_absent."""
+        seen = set()
+        for factory in entries:
+            seen.add(factory.key)
+            if factory.key == ('or',):
+                self.assertEqual('absent', factory.storage_kind)
+                self.assertEqual(None, factory.sha1)
+                self.assertEqual(None, factory.parents)
+            else:
+                self.assertValidStorageKind(factory.storage_kind)
+                self.assertEqual(f.get_sha1s([factory.key[0]])[0], factory.sha1)
+                self.assertEqual(parents[factory.key[0]], factory.parents)
+                self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
+                    str)
+        self.assertEqual(
+            set([('base',), ('left',), ('right',), ('merged',), ('or',)]),
+            seen)
+
+    def test_filter_absent_records(self):
+        """Requested missing records can be filter trivially."""
+        f, parents = get_diamond_vf(self.get_file())
+        entries = f.get_record_stream(['merged', 'left', 'right', 'extra', 'base'],
+            'unordered', False)
+        seen = set()
+        self.capture_stream(f, versionedfile.filter_absent(entries), seen.add,
+            parents)
+        self.assertEqual(set([('base',), ('left',), ('right',), ('merged',)]),
+            seen)
+
+    def test_insert_record_stream_empty(self):
+        """Inserting an empty record stream should work."""
+        f = self.get_file()
+        stream = []
+        f.insert_record_stream([])
+
+    def assertIdenticalVersionedFile(self, left, right):
+        """Assert that left and right have the same contents."""
+        self.assertEqual(set(left.versions()), set(right.versions()))
+        self.assertEqual(left.get_parent_map(left.versions()),
+            right.get_parent_map(right.versions()))
+        for v in left.versions():
+            self.assertEqual(left.get_text(v), right.get_text(v))
+
+    def test_insert_record_stream_fulltexts(self):
+        """Any file should accept a stream of fulltexts."""
+        f = self.get_file()
+        weave_vf = WeaveFile('source', get_transport(self.get_url('.')),
+            create=True, get_scope=self.get_transaction)
+        source, _ = get_diamond_vf(weave_vf)
+        stream = source.get_record_stream(source.versions(), 'topological',
+            False)
+        f.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(f, source)
+
+    def test_insert_record_stream_fulltexts_noeol(self):
+        """Any file should accept a stream of fulltexts."""
+        f = self.get_file()
+        weave_vf = WeaveFile('source', get_transport(self.get_url('.')),
+            create=True, get_scope=self.get_transaction)
+        source, _ = get_diamond_vf(weave_vf, trailing_eol=False)
+        stream = source.get_record_stream(source.versions(), 'topological',
+            False)
+        f.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(f, source)
+
+    def test_insert_record_stream_annotated_knits(self):
+        """Any file should accept a stream from plain knits."""
+        f = self.get_file()
+        source = make_file_knit('source', get_transport(self.get_url('.')),
+            create=True)
+        get_diamond_vf(source)
+        stream = source.get_record_stream(source.versions(), 'topological',
+            False)
+        f.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(f, source)
+
+    def test_insert_record_stream_annotated_knits_noeol(self):
+        """Any file should accept a stream from plain knits."""
+        f = self.get_file()
+        source = make_file_knit('source', get_transport(self.get_url('.')),
+            create=True)
+        get_diamond_vf(source, trailing_eol=False)
+        stream = source.get_record_stream(source.versions(), 'topological',
+            False)
+        f.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(f, source)
+
+    def test_insert_record_stream_plain_knits(self):
+        """Any file should accept a stream from plain knits."""
+        f = self.get_file()
+        source = make_file_knit('source', get_transport(self.get_url('.')),
+            create=True, factory=KnitPlainFactory())
+        get_diamond_vf(source)
+        stream = source.get_record_stream(source.versions(), 'topological',
+            False)
+        f.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(f, source)
+
+    def test_insert_record_stream_plain_knits_noeol(self):
+        """Any file should accept a stream from plain knits."""
+        f = self.get_file()
+        source = make_file_knit('source', get_transport(self.get_url('.')),
+            create=True, factory=KnitPlainFactory())
+        get_diamond_vf(source, trailing_eol=False)
+        stream = source.get_record_stream(source.versions(), 'topological',
+            False)
+        f.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(f, source)
+
+    def test_insert_record_stream_existing_keys(self):
+        """Inserting keys already in a file should not error."""
+        f = self.get_file()
+        source = make_file_knit('source', get_transport(self.get_url('.')),
+            create=True, factory=KnitPlainFactory())
+        get_diamond_vf(source)
+        # insert some keys into f.
+        get_diamond_vf(f, left_only=True)
+        stream = source.get_record_stream(source.versions(), 'topological',
+            False)
+        f.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(f, source)
+
+    def test_insert_record_stream_missing_keys(self):
+        """Inserting a stream with absent keys should raise an error."""
+        f = self.get_file()
+        source = make_file_knit('source', get_transport(self.get_url('.')),
+            create=True, factory=KnitPlainFactory())
+        stream = source.get_record_stream(['missing'], 'topological',
+            False)
+        self.assertRaises(errors.RevisionNotPresent, f.insert_record_stream,
+            stream)
+
+    def test_insert_record_stream_out_of_order(self):
+        """An out of order stream can either error or work."""
+        f, parents = get_diamond_vf(self.get_file())
+        origin_entries = f.get_record_stream(['origin'], 'unordered', False)
+        end_entries = f.get_record_stream(['merged', 'left'],
+            'topological', False)
+        start_entries = f.get_record_stream(['right', 'base'],
+            'topological', False)
+        entries = chain(origin_entries, end_entries, start_entries)
+        target = self.get_file('target')
+        try:
+            target.insert_record_stream(entries)
+        except RevisionNotPresent:
+            # Must not have corrupted the file.
+            target.check()
+        else:
+            self.assertIdenticalVersionedFile(f, target)
+
+    def test_insert_record_stream_delta_missing_basis_no_corruption(self):
+        """Insertion where a needed basis is not included aborts safely."""
+        # Annotated source - deltas can be used in any knit.
+        source = make_file_knit('source', get_transport(self.get_url('.')),
+            create=True)
+        get_diamond_vf(source)
+        entries = source.get_record_stream(['origin', 'merged'], 'unordered', False)
+        f = self.get_file()
+        self.assertRaises(RevisionNotPresent, f.insert_record_stream, entries)
+        f.check()
+        self.assertFalse(f.has_version('merged'))
+
     def test_adds_with_parent_texts(self):
         f = self.get_file()
         parent_texts = {}
@@ -358,7 +649,8 @@
         self._transaction = 'after'
         self.assertRaises(errors.OutSideTransaction, f.add_lines, '', [], [])
         self.assertRaises(errors.OutSideTransaction, f.add_lines_with_ghosts, '', [], [])
-        self.assertRaises(errors.OutSideTransaction, f.join, '')
+        self.assertRaises(errors.OutSideTransaction, self.applyDeprecated,
+            one_five, f.join, '')
         
     def test_clone_text(self):
         f = self.get_file()
@@ -693,7 +985,8 @@
                           'base',
                           [],
                           [])
-        self.assertRaises(errors.ReadOnlyError, vf.join, 'base')
+        self.assertRaises(errors.ReadOnlyError, self.applyDeprecated, one_five,
+            vf.join, 'base')
     
     def test_get_sha1s(self):
         # check the sha1 data is available
@@ -778,8 +1071,8 @@
 
 class TestKnit(TestCaseWithMemoryTransport, VersionedFileTestMixIn):
 
-    def get_file(self, name='foo'):
-        return self.get_factory()(name, get_transport(self.get_url('.')),
+    def get_file(self, name='foo', create=True):
+        return make_file_knit(name, get_transport(self.get_url('.')),
             delta=True, create=True, get_scope=self.get_transaction)
 
     def get_factory(self):
@@ -792,9 +1085,7 @@
         return knit
 
     def reopen_file(self, name='foo', create=False):
-        return self.get_factory()(name, get_transport(self.get_url('.')),
-            delta=True,
-            create=create)
+        return self.get_file(name, create)
 
     def test_detection(self):
         knit = self.get_file()
@@ -808,8 +1099,10 @@
 class TestPlaintextKnit(TestKnit):
     """Test a knit with no cached annotations"""
 
-    def get_factory(self):
-        return make_file_knit
+    def get_file(self, name='foo', create=True):
+        return make_file_knit(name, get_transport(self.get_url('.')),
+            delta=True, create=create, get_scope=self.get_transaction,
+            factory=_mod_knit.KnitPlainFactory())
 
 
 class TestPlanMergeVersionedFile(TestCaseWithMemoryTransport):
@@ -1243,3 +1536,147 @@
 
     overlappedInsertExpected = ['aaa', '<<<<<<< ', 'xxx', 'yyy', '=======', 
                                 'xxx', '>>>>>>> ', 'bbb']
+
+
+class TestContentFactoryAdaption(TestCaseWithMemoryTransport):
+
+    def test_select_adaptor(self):
+        """Test expected adapters exist."""
+        # One scenario for each lookup combination we expect to use.
+        # Each is source_kind, requested_kind, adapter class
+        scenarios = [
+            ('knit-delta-gz', 'fulltext', _mod_knit.DeltaPlainToFullText),
+            ('knit-ft-gz', 'fulltext', _mod_knit.FTPlainToFullText),
+            ('knit-annotated-delta-gz', 'knit-delta-gz',
+                _mod_knit.DeltaAnnotatedToUnannotated),
+            ('knit-annotated-delta-gz', 'fulltext',
+                _mod_knit.DeltaAnnotatedToFullText),
+            ('knit-annotated-ft-gz', 'knit-ft-gz',
+                _mod_knit.FTAnnotatedToUnannotated),
+            ('knit-annotated-ft-gz', 'fulltext',
+                _mod_knit.FTAnnotatedToFullText),
+            ]
+        for source, requested, klass in scenarios:
+            adapter_factory = versionedfile.adapter_registry.get(
+                (source, requested))
+            adapter = adapter_factory(None)
+            self.assertIsInstance(adapter, klass)
+
+    def get_knit(self, annotated=True):
+        if annotated:
+            factory = KnitAnnotateFactory()
+        else:
+            factory = KnitPlainFactory()
+        return make_file_knit('knit', self.get_transport('.'), delta=True,
+            create=True, factory=factory)
+
+    def helpGetBytes(self, f, ft_adapter, delta_adapter):
+        """Grab the interested adapted texts for tests."""
+        # origin is a fulltext
+        entries = f.get_record_stream(['origin'], 'unordered', False)
+        base = entries.next()
+        ft_data = ft_adapter.get_bytes(base, base.get_bytes_as(base.storage_kind))
+        # merged is both a delta and multiple parents.
+        entries = f.get_record_stream(['merged'], 'unordered', False)
+        merged = entries.next()
+        delta_data = delta_adapter.get_bytes(merged,
+            merged.get_bytes_as(merged.storage_kind))
+        return ft_data, delta_data
+
+    def test_deannotation_noeol(self):
+        """Test converting annotated knits to unannotated knits."""
+        # we need a full text, and a delta
+        f, parents = get_diamond_vf(self.get_knit(), trailing_eol=False)
+        ft_data, delta_data = self.helpGetBytes(f,
+            _mod_knit.FTAnnotatedToUnannotated(None),
+            _mod_knit.DeltaAnnotatedToUnannotated(None))
+        self.assertEqual(
+            'version origin 1 b284f94827db1fa2970d9e2014f080413b547a7e\n'
+            'origin\n'
+            'end origin\n',
+            GzipFile(mode='rb', fileobj=StringIO(ft_data)).read())
+        self.assertEqual(
+            'version merged 4 32c2e79763b3f90e8ccde37f9710b6629c25a796\n'
+            '1,2,3\nleft\nright\nmerged\nend merged\n',
+            GzipFile(mode='rb', fileobj=StringIO(delta_data)).read())
+
+    def test_deannotation(self):
+        """Test converting annotated knits to unannotated knits."""
+        # we need a full text, and a delta
+        f, parents = get_diamond_vf(self.get_knit())
+        ft_data, delta_data = self.helpGetBytes(f,
+            _mod_knit.FTAnnotatedToUnannotated(None),
+            _mod_knit.DeltaAnnotatedToUnannotated(None))
+        self.assertEqual(
+            'version origin 1 00e364d235126be43292ab09cb4686cf703ddc17\n'
+            'origin\n'
+            'end origin\n',
+            GzipFile(mode='rb', fileobj=StringIO(ft_data)).read())
+        self.assertEqual(
+            'version merged 3 ed8bce375198ea62444dc71952b22cfc2b09226d\n'
+            '2,2,2\nright\nmerged\nend merged\n',
+            GzipFile(mode='rb', fileobj=StringIO(delta_data)).read())
+
+    def test_annotated_to_fulltext_no_eol(self):
+        """Test adapting annotated knits to full texts (for -> weaves)."""
+        # we need a full text, and a delta
+        f, parents = get_diamond_vf(self.get_knit(), trailing_eol=False)
+        # Reconstructing a full text requires a backing versioned file, and it
+        # must have the base lines requested from it.
+        logged_vf = versionedfile.RecordingVersionedFileDecorator(f)
+        ft_data, delta_data = self.helpGetBytes(f,
+            _mod_knit.FTAnnotatedToFullText(None),
+            _mod_knit.DeltaAnnotatedToFullText(logged_vf))
+        self.assertEqual('origin', ft_data)
+        self.assertEqual('base\nleft\nright\nmerged', delta_data)
+        self.assertEqual([('get_lines', 'left')], logged_vf.calls)
+
+    def test_annotated_to_fulltext(self):
+        """Test adapting annotated knits to full texts (for -> weaves)."""
+        # we need a full text, and a delta
+        f, parents = get_diamond_vf(self.get_knit())
+        # Reconstructing a full text requires a backing versioned file, and it
+        # must have the base lines requested from it.
+        logged_vf = versionedfile.RecordingVersionedFileDecorator(f)
+        ft_data, delta_data = self.helpGetBytes(f,
+            _mod_knit.FTAnnotatedToFullText(None),
+            _mod_knit.DeltaAnnotatedToFullText(logged_vf))
+        self.assertEqual('origin\n', ft_data)
+        self.assertEqual('base\nleft\nright\nmerged\n', delta_data)
+        self.assertEqual([('get_lines', 'left')], logged_vf.calls)
+
+    def test_unannotated_to_fulltext(self):
+        """Test adapting unannotated knits to full texts.
+        
+        This is used for -> weaves, and for -> annotated knits.
+        """
+        # we need a full text, and a delta
+        f, parents = get_diamond_vf(self.get_knit(annotated=False))
+        # Reconstructing a full text requires a backing versioned file, and it
+        # must have the base lines requested from it.
+        logged_vf = versionedfile.RecordingVersionedFileDecorator(f)
+        ft_data, delta_data = self.helpGetBytes(f,
+            _mod_knit.FTPlainToFullText(None),
+            _mod_knit.DeltaPlainToFullText(logged_vf))
+        self.assertEqual('origin\n', ft_data)
+        self.assertEqual('base\nleft\nright\nmerged\n', delta_data)
+        self.assertEqual([('get_lines', 'left')], logged_vf.calls)
+
+    def test_unannotated_to_fulltext_no_eol(self):
+        """Test adapting unannotated knits to full texts.
+        
+        This is used for -> weaves, and for -> annotated knits.
+        """
+        # we need a full text, and a delta
+        f, parents = get_diamond_vf(self.get_knit(annotated=False),
+            trailing_eol=False)
+        # Reconstructing a full text requires a backing versioned file, and it
+        # must have the base lines requested from it.
+        logged_vf = versionedfile.RecordingVersionedFileDecorator(f)
+        ft_data, delta_data = self.helpGetBytes(f,
+            _mod_knit.FTPlainToFullText(None),
+            _mod_knit.DeltaPlainToFullText(logged_vf))
+        self.assertEqual('origin', ft_data)
+        self.assertEqual('base\nleft\nright\nmerged', delta_data)
+        self.assertEqual([('get_lines', 'left')], logged_vf.calls)
+

=== modified file 'bzrlib/tests/test_weave.py'
--- bzrlib/tests/test_weave.py	2008-04-08 03:39:43 +0000
+++ bzrlib/tests/test_weave.py	2008-04-29 05:27:08 +0000
@@ -657,66 +657,6 @@
         self.weave1.add_lines('v1', [], self.lines1)
         self.weave1.add_lines('v2', ['v1'], ['hello\n', 'world\n'])
         self.weave1.add_lines('v3', ['v2'], self.lines3)
-        
-    def test_join_empty(self):
-        """Join two empty weaves."""
-        eq = self.assertEqual
-        w1 = Weave()
-        w2 = Weave()
-        w1.join(w2)
-        eq(len(w1), 0)
-        
-    def test_join_empty_to_nonempty(self):
-        """Join empty weave onto nonempty."""
-        self.weave1.join(Weave())
-        self.assertEqual(len(self.weave1), 3)
-
-    def test_join_unrelated(self):
-        """Join two weaves with no history in common."""
-        wb = Weave()
-        wb.add_lines('b1', [], ['line from b\n'])
-        w1 = self.weave1
-        w1.join(wb)
-        eq = self.assertEqual
-        eq(len(w1), 4)
-        eq(sorted(w1.versions()),
-           ['b1', 'v1', 'v2', 'v3'])
-
-    def test_join_related(self):
-        wa = self.weave1.copy()
-        wb = self.weave1.copy()
-        wa.add_lines('a1', ['v3'], ['hello\n', 'sweet\n', 'world\n'])
-        wb.add_lines('b1', ['v3'], ['hello\n', 'pale blue\n', 'world\n'])
-        eq = self.assertEquals
-        eq(len(wa), 4)
-        eq(len(wb), 4)
-        wa.join(wb)
-        eq(len(wa), 5)
-        eq(wa.get_lines('b1'),
-           ['hello\n', 'pale blue\n', 'world\n'])
-
-    def test_join_text_disagreement(self):
-        """Cannot join weaves with different texts for a version."""
-        wa = Weave()
-        wb = Weave()
-        wa.add_lines('v1', [], ['hello\n'])
-        wb.add_lines('v1', [], ['not\n', 'hello\n'])
-        self.assertRaises(WeaveError,
-                          wa.join, wb)
-
-    def test_join_unordered(self):
-        """Join weaves where indexes differ.
-        
-        The source weave contains a different version at index 0."""
-        wa = self.weave1.copy()
-        wb = Weave()
-        wb.add_lines('x1', [], ['line from x1\n'])
-        wb.add_lines('v1', [], ['hello\n'])
-        wb.add_lines('v2', ['v1'], ['hello\n', 'world\n'])
-        wa.join(wb)
-        eq = self.assertEquals
-        eq(sorted(wa.versions()), ['v1', 'v2', 'v3', 'x1',])
-        eq(wa.get_text('x1'), 'line from x1\n')
 
     def test_written_detection(self):
         # Test detection of weave file corruption.
@@ -779,57 +719,6 @@
         return Weave._extract(self, versions)
 
 
-class JoinOptimization(TestCase):
-    """Test that Weave.join() doesn't extract all texts, only what must be done."""
-
-    def test_join(self):
-        w1 = InstrumentedWeave()
-        w2 = InstrumentedWeave()
-
-        txt0 = ['a\n']
-        txt1 = ['a\n', 'b\n']
-        txt2 = ['a\n', 'c\n']
-        txt3 = ['a\n', 'b\n', 'c\n']
-
-        w1.add_lines('txt0', [], txt0) # extract 1a
-        w2.add_lines('txt0', [], txt0) # extract 1b
-        w1.add_lines('txt1', ['txt0'], txt1)# extract 2a
-        w2.add_lines('txt2', ['txt0'], txt2)# extract 2b
-        w1.join(w2) # extract 3a to add txt2 
-        w2.join(w1) # extract 3b to add txt1 
-
-        w1.add_lines('txt3', ['txt1', 'txt2'], txt3) # extract 4a 
-        w2.add_lines('txt3', ['txt2', 'txt1'], txt3) # extract 4b
-        # These secretly have inverted parents
-
-        # This should not have to do any extractions
-        w1.join(w2) # NO extract, texts already present with same parents
-        w2.join(w1) # NO extract, texts already present with same parents
-
-        self.assertEqual(4, w1._extract_count)
-        self.assertEqual(4, w2._extract_count)
-
-    def test_double_parent(self):
-        # It should not be considered illegal to add
-        # a revision with the same parent twice
-        w1 = InstrumentedWeave()
-        w2 = InstrumentedWeave()
-
-        txt0 = ['a\n']
-        txt1 = ['a\n', 'b\n']
-        txt2 = ['a\n', 'c\n']
-        txt3 = ['a\n', 'b\n', 'c\n']
-
-        w1.add_lines('txt0', [], txt0)
-        w2.add_lines('txt0', [], txt0)
-        w1.add_lines('txt1', ['txt0'], txt1)
-        w2.add_lines('txt1', ['txt0', 'txt0'], txt1)
-        # Same text, effectively the same, because the
-        # parent is only repeated
-        w1.join(w2) # extract 3a to add txt2 
-        w2.join(w1) # extract 3b to add txt1 
-
-
 class TestNeedsReweave(TestCase):
     """Internal corner cases for when reweave is needed."""
 

=== modified file 'bzrlib/versionedfile.py'
--- bzrlib/versionedfile.py	2008-04-24 07:22:53 +0000
+++ bzrlib/versionedfile.py	2008-05-12 01:23:41 +0000
@@ -37,10 +37,74 @@
 from cStringIO import StringIO
 
 from bzrlib.inter import InterObject
+from bzrlib.registry import Registry
 from bzrlib.symbol_versioning import *
 from bzrlib.textmerge import TextMerge
 
 
+adapter_registry = Registry()
+adapter_registry.register_lazy(('knit-delta-gz', 'fulltext'), 'bzrlib.knit',
+    'DeltaPlainToFullText')
+adapter_registry.register_lazy(('knit-ft-gz', 'fulltext'), 'bzrlib.knit',
+    'FTPlainToFullText')
+adapter_registry.register_lazy(('knit-annotated-delta-gz', 'knit-delta-gz'),
+    'bzrlib.knit', 'DeltaAnnotatedToUnannotated')
+adapter_registry.register_lazy(('knit-annotated-delta-gz', 'fulltext'),
+    'bzrlib.knit', 'DeltaAnnotatedToFullText')
+adapter_registry.register_lazy(('knit-annotated-ft-gz', 'knit-ft-gz'),
+    'bzrlib.knit', 'FTAnnotatedToUnannotated')
+adapter_registry.register_lazy(('knit-annotated-ft-gz', 'fulltext'),
+    'bzrlib.knit', 'FTAnnotatedToFullText')
+
+
+class ContentFactory(object):
+    """Abstract interface for insertion and retrieval from a VersionedFile.
+    
+    :ivar sha1: None, or the sha1 of the content fulltext.
+    :ivar storage_kind: The native storage kind of this factory. One of
+        'mpdiff', 'knit-annotated-ft', 'knit-annotated-delta', 'knit-ft',
+        'knit-delta', 'fulltext', 'knit-annotated-ft-gz',
+        'knit-annotated-delta-gz', 'knit-ft-gz', 'knit-delta-gz'.
+    :ivar key: The key of this content. Each key is a tuple with a single
+        string in it.
+    :ivar parents: A tuple of parent keys for self.key. If the object has
+        no parent information, None (as opposed to () for an empty list of
+        parents).
+        """
+
+    def __init__(self):
+        """Create a ContentFactory."""
+        self.sha1 = None
+        self.storage_kind = None
+        self.key = None
+        self.parents = None
+
+
+class AbsentContentFactory(object):
+    """A placeholder content factory for unavailable texts.
+    
+    :ivar sha1: None.
+    :ivar storage_kind: 'absent'.
+    :ivar key: The key of this content. Each key is a tuple with a single
+        string in it.
+    :ivar parents: None.
+    """
+
+    def __init__(self, key):
+        """Create a ContentFactory."""
+        self.sha1 = None
+        self.storage_kind = 'absent'
+        self.key = key
+        self.parents = None
+
+
+def filter_absent(record_stream):
+    """Adapt a record stream to remove absent records."""
+    for record in record_stream:
+        if record.storage_kind != 'absent':
+            yield record
+
+
 class VersionedFile(object):
     """Versioned text file storage.
     
@@ -63,9 +127,22 @@
         """Copy this versioned file to name on transport."""
         raise NotImplementedError(self.copy_to)
 
-    def versions(self):
-        """Return a unsorted list of versions."""
-        raise NotImplementedError(self.versions)
+    def get_record_stream(self, versions, ordering, include_delta_closure):
+        """Get a stream of records for versions.
+
+        :param versions: The versions to include. Each version is a tuple
+            (version,).
+        :param ordering: Either 'unordered' or 'topological'. A topologically
+            sorted stream has compression parents strictly before their
+            children.
+        :param include_delta_closure: If True then the closure across any
+            compression parents will be included (in the data content of the
+            stream, not in the emitted records). This guarantees that
+            'fulltext' can be used successfully on every record.
+        :return: An iterator of ContentFactory objects, each of which is only
+            valid until the iterator is advanced.
+        """
+        raise NotImplementedError(self.get_record_stream)
 
     @deprecated_method(one_four)
     def has_ghost(self, version_id):
@@ -76,6 +153,15 @@
         """Returns whether version is present."""
         raise NotImplementedError(self.has_version)
 
+    def insert_record_stream(self, stream):
+        """Insert a record stream into this versioned file.
+
+        :param stream: A stream of records to insert. 
+        :return: None
+        :seealso VersionedFile.get_record_stream:
+        """
+        raise NotImplementedError
+
     def add_lines(self, version_id, parents, lines, parent_texts=None,
         left_matching_blocks=None, nostore_sha=None, random_id=False,
         check_content=True):
@@ -440,6 +526,7 @@
         """
         raise NotImplementedError(self.annotate)
 
+    @deprecated_method(one_five)
     def join(self, other, pb=None, msg=None, version_ids=None,
              ignore_missing=False):
         """Integrate versions from other into this versioned file.
@@ -519,6 +606,28 @@
         return PlanWeaveMerge(plan, a_marker, b_marker).merge_lines()[0]
 
 
+class RecordingVersionedFileDecorator(object):
+    """A minimal versioned file that records calls made on it.
+    
+    Only enough methods have been added to support tests using it to date.
+
+    :ivar calls: A list of the calls made; can be reset at any time by
+        assigning [] to it.
+    """
+
+    def __init__(self, backing_vf):
+        """Create a RecordingVersionedFileDecorator decorating backing_vf.
+        
+        :param backing_vf: The versioned file to answer all methods.
+        """
+        self._backing_vf = backing_vf
+        self.calls = []
+
+    def get_lines(self, version_ids):
+        self.calls.append(("get_lines", version_ids))
+        return self._backing_vf.get_lines(version_ids)
+
+
 class _PlanMergeVersionedFile(object):
     """A VersionedFile for uncommitted and committed texts.
 

=== modified file 'bzrlib/weave.py'
--- bzrlib/weave.py	2008-04-24 07:22:53 +0000
+++ bzrlib/weave.py	2008-05-11 23:49:50 +0000
@@ -78,21 +78,52 @@
 from bzrlib import (
     progress,
     )
-from bzrlib.trace import mutter
 from bzrlib.errors import (WeaveError, WeaveFormatError, WeaveParentMismatch,
         RevisionAlreadyPresent,
         RevisionNotPresent,
+        UnavailableRepresentation,
         WeaveRevisionAlreadyPresent,
         WeaveRevisionNotPresent,
         )
 import bzrlib.errors as errors
-from bzrlib.osutils import sha_strings
+from bzrlib.osutils import sha_strings, split_lines
 import bzrlib.patiencediff
+from bzrlib.symbol_versioning import *
+from bzrlib.trace import mutter
 from bzrlib.tsort import topo_sort
-from bzrlib.versionedfile import VersionedFile, InterVersionedFile
+from bzrlib.versionedfile import (
+    AbsentContentFactory,
+    adapter_registry,
+    ContentFactory,
+    InterVersionedFile,
+    VersionedFile,
+    )
 from bzrlib.weavefile import _read_weave_v5, write_weave_v5
 
 
+class WeaveContentFactory(ContentFactory):
+    """Content factory for streaming from weaves.
+
+    :seealso ContentFactory:
+    """
+
+    def __init__(self, version, weave):
+        """Create a WeaveContentFactory for version from weave."""
+        ContentFactory.__init__(self)
+        self.sha1 = weave.get_sha1s([version])[0]
+        self.key = (version,)
+        parents = weave.get_parent_map([version])[version]
+        self.parents = tuple((parent,) for parent in parents)
+        self.storage_kind = 'fulltext'
+        self._weave = weave
+
+    def get_bytes_as(self, storage_kind):
+        if storage_kind == 'fulltext':
+            return self._weave.get_text(self.key[0])
+        else:
+            raise UnavailableRepresentation(self.key, storage_kind, 'fulltext')
+
+
 class Weave(VersionedFile):
     """weave - versioned text file storage.
     
@@ -263,6 +294,30 @@
 
     __contains__ = has_version
 
+    def get_record_stream(self, versions, ordering, include_delta_closure):
+        """Get a stream of records for versions.
+
+        :param versions: The versions to include. Each version is a tuple
+            (version,).
+        :param ordering: Either 'unordered' or 'topological'. A topologically
+            sorted stream has compression parents strictly before their
+            children.
+        :param include_delta_closure: If True then the closure across any
+            compression parents will be included (in the opaque data).
+        :return: An iterator of ContentFactory objects, each of which is only
+            valid until the iterator is advanced.
+        """
+        if ordering == 'topological':
+            parents = self.get_parent_map(versions)
+            new_versions = topo_sort(parents)
+            new_versions.extend(set(versions).difference(set(parents)))
+            versions = new_versions
+        for version in versions:
+            if version in self:
+                yield WeaveContentFactory(version, self)
+            else:
+                yield AbsentContentFactory((version,))
+
     def get_parent_map(self, version_ids):
         """See VersionedFile.get_parent_map."""
         result = {}
@@ -277,6 +332,38 @@
     def get_parents_with_ghosts(self, version_id):
         raise NotImplementedError(self.get_parents_with_ghosts)
 
+    def insert_record_stream(self, stream):
+        """Insert a record stream into this versioned file.
+
+        :param stream: A stream of records to insert. 
+        :return: None
+        :seealso VersionedFile.get_record_stream:
+        """
+        adapters = {}
+        for record in stream:
+            # Raise an error when a record is missing.
+            if record.storage_kind == 'absent':
+                raise RevisionNotPresent([record.key[0]], self)
+            # adapt to non-tuple interface
+            parents = [parent[0] for parent in record.parents]
+            if record.storage_kind == 'fulltext':
+                self.add_lines(record.key[0], parents,
+                    split_lines(record.get_bytes_as('fulltext')))
+            else:
+                adapter_key = record.storage_kind, 'fulltext'
+                try:
+                    adapter = adapters[adapter_key]
+                except KeyError:
+                    adapter_factory = adapter_registry.get(adapter_key)
+                    adapter = adapter_factory(self)
+                    adapters[adapter_key] = adapter
+                lines = split_lines(adapter.get_bytes(
+                    record, record.get_bytes_as(record.storage_kind)))
+                try:
+                    self.add_lines(record.key[0], parents, lines)
+                except RevisionAlreadyPresent:
+                    pass
+
     def _check_repeated_add(self, name, parents, text, sha1):
         """Check that a duplicated add is OK.
 
@@ -911,6 +998,11 @@
         """See VersionedFile.get_suffixes()."""
         return [WeaveFile.WEAVE_SUFFIX]
 
+    def insert_record_stream(self, stream):
+        super(WeaveFile, self).insert_record_stream(stream)
+        self._save()
+
+    @deprecated_method(one_five)
     def join(self, other, pb=None, msg=None, version_ids=None,
              ignore_missing=False):
         """Join other into self and save."""

=== modified file 'doc/developers/index.txt'
--- doc/developers/index.txt	2008-05-09 09:52:59 +0000
+++ doc/developers/index.txt	2008-05-11 23:49:50 +0000
@@ -39,6 +39,9 @@
 * `Container format <container-format.html>`_ |--| Notes on a container format
   for streaming and storing Bazaar data.
 
+* `Repository stream <repository-stream.html>`_ |--| Notes on streaming data
+  for repositories (a layer above the container format).
+
 * `Indices <indices.html>`_ |--| The index facilities available within bzrlib.
 
 * `Inventories <inventory.html>`_ |--| Tree shape abstraction.

=== added file 'doc/developers/repository-stream.txt'
--- doc/developers/repository-stream.txt	1970-01-01 00:00:00 +0000
+++ doc/developers/repository-stream.txt	2008-04-16 04:07:40 +0000
@@ -0,0 +1,200 @@
+==================
+Repository Streams
+==================
+
+Status
+======
+
+:Date: 2008-04-11
+
+This document describes the proposed programming interface for streaming
+data from and into repositories. This programming interface should allow
+a single interface for pulling data from and inserting data into a Bazaar
+repository.
+
+.. contents::
+
+
+Motivation
+==========
+
+To eliminate the current requirement that extracting data from a
+repository requires either using a slow format, or knowing the format of
+both the source repository and the target repository.
+
+
+Use Cases
+=========
+
+Here's a brief description of use cases this interface is intended to
+support.
+
+Fetch operations
+----------------
+
+We fetch data between repositories as part of push/pull/branch operations.
+Fetching data is currently an very interactive process with lots of
+requests. For performance having the data be supplied in a stream will
+improve push and pull to remote servers. For purely local operations the
+streaming logic should help reduce memory pressure. In fetch operations
+we always know the formats of both the source and target.
+
+Smart server operations
+~~~~~~~~~~~~~~~~~~~~~~~
+
+With the smart server we support one streaming format, but this is only
+usable when both the client and server have the same model of data, and
+requires non-optimal IO ordering for pack to pack operations. Ideally we
+can both provide optimal IO ordering the pack to pack case, and correct
+ordering for pack to knits.
+
+Bundles
+-------
+
+Bundles also create a stream of data for revisions from a repository.
+Unlike fetch operations we do not know the format of the target at the
+time the stream is created. It would be good to be able to treat bundles
+as frozen branches and repositories, so a serialised stream should be
+suitable for this.
+
+Data conversion
+---------------
+
+At this point we are not trying to integrate data conversion into this
+interface, though it is likely possible.
+
+
+Characteristics
+===============
+
+Some key aspects of the described interface are discussed in this section.
+
+Single round trip
+-----------------
+
+All users of this should be able to create an appropriate stream from a
+single round trip.
+
+Forward-only reads
+------------------
+
+There should be no need to seek in a stream when inserting data from it
+into a repository. This places an ordering constraint on streams which
+some repositories do not need.
+
+
+Serialisation
+=============
+
+At this point serialisation of a repository stream has not been specified.
+Some considerations to bear in mind about serialisation are worth noting
+however.
+
+Weaves
+------
+
+While there shouldn't be too many users of weave repositories anymore,
+avoiding pathological behaviour when a weave is being read is a good idea.
+Having the weave itself embedded in the stream is very straight forward
+and does not need expensive on the fly extraction and re-diffing to take
+place.
+
+Bundles
+-------
+
+Being able to perform random reads from a repository stream which is a
+bundle would allow stacking a bundle and a real repository together. This
+will need the pack container format to be used in such a way that we can
+avoid reading more data than needed within the pack container's readv
+interface.
+
+
+Specification
+=============
+
+This describes the interface for requesting a stream, and the programming
+interface a stream must provide. Streams that have been serialised should
+expose the same interface.
+
+Requesting a stream
+-------------------
+
+To request a stream, three parameters are needed:
+
+ * A revision search to select the revisions to include.
+ * A data ordering flag. There are two values for this - 'unordered' and
+   'topological'. 'unordered' streams are useful when inserting into
+   repositories that have the ability to perform atomic insertions.
+   'topological' streams are useful when converting data, or when
+   inserting into repositories that cannot perform atomic insertions (such
+   as knit or weave based repositories).
+ * A complete_inventory flag. When provided this flag signals the stream
+   generator to include all the data needed to construct the inventory of
+   each revision included in the stream, rather than just deltas. This is
+   useful when converting data from a repository with a different
+   inventory serialisation, as pure deltas would not be able to be
+   reconstructed.
+
+
+Structure of a stream
+---------------------
+
+A stream is an object. It can be consistency checked via the ``check``
+method (which consumes the stream). The ``iter_contents`` method can be
+used to iterate the contents of the stream. The contents of the stream are
+a series of top level records, each of which contains one or more
+bytestrings (potentially as a delta against another item in the
+repository) and some optional metadata.
+
+
+Consuming a stream
+------------------
+
+To consume a stream, obtain an iterator from the streams
+``iter_contents`` method. This iterator will yield the top level records.
+Each record has two attributes. One is ``key_prefix`` which is a tuple key
+prefix for the names of each of the bytestrings in the record. The other
+attribute is ``entries``, an iterator of the individual items in the
+record. Each item that the iterator yields is a factory which has metadata
+about the entry and the ability to return the compressed bytes. This
+factory can be decorated to allow obtaining different representations (for
+example from a compressed knit fulltext to a plain fulltext). 
+
+In pseudocode::
+
+  stream = repository.get_repository_stream(search, UNORDERED, False)
+  for record in stream.iter_contents():
+      for factory in record.entries:
+          compression = factory.storage_kind
+          print "Object %s, compression type %s, %d bytes long." % (
+              record.key_prefix + factory.key,
+              compression, len(factory.get_bytes_as(compression)))
+
+This structure should allow stream adapters to be written which can coerce
+all records to the type of compression that a particular client needs. For
+instance, inserting into weaves requires fulltexts, so a stream would be
+adapted for weaves by an adapter that takes a stream, and the target
+weave, and then uses the target weave to reconstruct full texts (which is
+all that the weave inserter would ask for). In a similar approach, a
+stream could internally delta compress many fulltexts and be able to
+answer both fulltext and compressed record requests without extra IO.
+
+factory metadata
+~~~~~~~~~~~~~~~~
+
+Valid attributes on the factory are:
+ * sha1: Optional ascii representation of the sha1 of the bytestring (after
+   delta reconstruction).
+ * storage_kind: Required kind of storage compression that has been used
+   on the bytestring. One of ``mpdiff``, ``knit-annotated-ft``,
+   ``knit-annotated-delta``, ``knit-ft``, ``knit-delta``, ``fulltext``.
+ * parents: Required graph parents to associate with this bytestring.
+ * compressor_data: Required opaque data relevant to the storage_kind.
+   (This is set to None when the compressor has no special state needed)
+ * key: The key for this bytestring. Like each parent this is a tuple that
+   should have the key_prefix prepended to it to give the unified
+   repository key name.
+
+..
+   vim: ft=rst tw=74 ai
+

